// TODO:
// Problemen:
// - frekwentie niet goed en afhankelijk van timer 1 sample frekwentie
//   Is misschien probleem met XOscillo, die lijkt er een factor 2 naast te zitten
// - seriele interface werkt niet omdat die D0 en D1 gebruikt
// $Id$
// Copyright (c) 2013 Ruud Mulder  All right reserved.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
#include "Arduino.h"
#include <avr/pgmspace.h>
// Use TimerOne library to set Timer 1 for generating wave samples
#include <TimerOne.h>
// PinChangeInt library is used to be able to generate interrupt via any digital input
#include <PinChangeInt.h>

// Wave generation is started immediately at the specified sample frequency.
// Samplefrequency (in Hz) can be up to 100kHz but must be less if
// more oscillators are used. Experiment what suits you best and is possible.
// For 16 bit fraction 1 bit is approx 0.00149 Hz
//   @ 100kHz sampling with a table of 1024.
// For 8 bit 'table' (pulse, triangle, sawtooth) this is a factor 4 higher
//   (approximately 0.00596).
// For a sample rate of 50kHz these value halve.
// N.B. the higher the sample frequency the easier it is to filter it out.
//===============================================
// Data definitions
//===============================================
#include "Oscillator.h"
//===============================================
// Global variables for interrupt routine
//===============================================
               float   SAMPLEFREQ   = 10000.0; // in Hz (can be adapted!)
const unsigned int     AMPLNORM     = 1000;    // = 1.0 (for normalizing)
const unsigned int     AMPLITUDEMAX = 127;
const unsigned int     SYNCPIN      = 12;       // pin that gets the sync pulse
const unsigned int     FANPIN       = 0;        // pin that gets analog fan input
volatile       char    outputvalue  = 0;       // -127..127
volatile       long    errortimeout = 0;       // the time the error is shown
volatile       boolean restart_wave = false;   // signals wave reset
const unsigned int   NUMOSC       = 3; // number of oscillators
// N.B. use functions to set values for oscarr.
volatile Oscillator  oscarr[NUMOSC];
#define DACOUT   (PORTD)
#define ERRORLED (13)
//===============================================
// Generate sine wave by table lookup
//===============================================
// Get 8 bit sine values based on a 1024 entries table (power of 2 for fast lookup).
// This ensures that the adjacent entries differ at most 1 from eachother.
// It consists solely of lookup functions.
// To generate a sine wave calculate the phase increment using
// frequency2phaseinc for the desired frequency and sample frequency
// and then getSine with the phase incremented by the calculated increment.
//
const unsigned int  PHASEFRACBITS = 16;     // use 16 bits as the fraction
const unsigned int  SINETABLESIZE = 1024;
const unsigned long SINETABLEMASK = 0x03FF; // only least significant 10 bits used
prog_char  sinetable[SINETABLESIZE] PROGMEM = {
    0,    1,    2,    2,    3,    4,    5,    5,
    6,    7,    8,    9,    9,   10,   11,   12,
   12,   13,   14,   15,   16,   16,   17,   18,
   19,   19,   20,   21,   22,   22,   23,   24,
   25,   26,   26,   27,   28,   29,   29,   30,
   31,   32,   32,   33,   34,   35,   35,   36,
   37,   38,   38,   39,   40,   41,   41,   42,
   43,   44,   44,   45,   46,   46,   47,   48,
   49,   49,   50,   51,   51,   52,   53,   54,
   54,   55,   56,   56,   57,   58,   58,   59,
   60,   61,   61,   62,   63,   63,   64,   65,
   65,   66,   67,   67,   68,   69,   69,   70,
   71,   71,   72,   72,   73,   74,   74,   75,
   76,   76,   77,   78,   78,   79,   79,   80,
   81,   81,   82,   82,   83,   84,   84,   85,
   85,   86,   86,   87,   88,   88,   89,   89,
   90,   90,   91,   91,   92,   93,   93,   94,
   94,   95,   95,   96,   96,   97,   97,   98,
   98,   99,   99,  100,  100,  101,  101,  102,
  102,  102,  103,  103,  104,  104,  105,  105,
  106,  106,  106,  107,  107,  108,  108,  109,
  109,  109,  110,  110,  111,  111,  111,  112,
  112,  112,  113,  113,  113,  114,  114,  114,
  115,  115,  115,  116,  116,  116,  117,  117,
  117,  118,  118,  118,  118,  119,  119,  119,
  120,  120,  120,  120,  121,  121,  121,  121,
  122,  122,  122,  122,  122,  123,  123,  123,
  123,  123,  124,  124,  124,  124,  124,  124,
  125,  125,  125,  125,  125,  125,  125,  126,
  126,  126,  126,  126,  126,  126,  126,  126,
  126,  126,  127,  127,  127,  127,  127,  127,
  127,  127,  127,  127,  127,  127,  127,  127,
  127,  127,  127,  127,  127,  127,  127,  127,
  127,  127,  127,  127,  127,  127,  127,  126,
  126,  126,  126,  126,  126,  126,  126,  126,
  126,  126,  125,  125,  125,  125,  125,  125,
  125,  124,  124,  124,  124,  124,  124,  123,
  123,  123,  123,  123,  122,  122,  122,  122,
  122,  121,  121,  121,  121,  120,  120,  120,
  120,  119,  119,  119,  118,  118,  118,  118,
  117,  117,  117,  116,  116,  116,  115,  115,
  115,  114,  114,  114,  113,  113,  113,  112,
  112,  112,  111,  111,  111,  110,  110,  109,
  109,  109,  108,  108,  107,  107,  106,  106,
  106,  105,  105,  104,  104,  103,  103,  102,
  102,  102,  101,  101,  100,  100,   99,   99,
   98,   98,   97,   97,   96,   96,   95,   95,
   94,   94,   93,   93,   92,   91,   91,   90,
   90,   89,   89,   88,   88,   87,   86,   86,
   85,   85,   84,   84,   83,   82,   82,   81,
   81,   80,   79,   79,   78,   78,   77,   76,
   76,   75,   74,   74,   73,   72,   72,   71,
   71,   70,   69,   69,   68,   67,   67,   66,
   65,   65,   64,   63,   63,   62,   61,   61,
   60,   59,   58,   58,   57,   56,   56,   55,
   54,   54,   53,   52,   51,   51,   50,   49,
   49,   48,   47,   46,   46,   45,   44,   44,
   43,   42,   41,   41,   40,   39,   38,   38,
   37,   36,   35,   35,   34,   33,   32,   32,
   31,   30,   29,   29,   28,   27,   26,   26,
   25,   24,   23,   22,   22,   21,   20,   19,
   19,   18,   17,   16,   16,   15,   14,   13,
   12,   12,   11,   10,    9,    9,    8,    7,
    6,    5,    5,    4,    3,    2,    2,    1,
    0,   -1,   -2,   -2,   -3,   -4,   -5,   -5,
   -6,   -7,   -8,   -9,   -9,  -10,  -11,  -12,
  -12,  -13,  -14,  -15,  -16,  -16,  -17,  -18,
  -19,  -19,  -20,  -21,  -22,  -22,  -23,  -24,
  -25,  -26,  -26,  -27,  -28,  -29,  -29,  -30,
  -31,  -32,  -32,  -33,  -34,  -35,  -35,  -36,
  -37,  -38,  -38,  -39,  -40,  -41,  -41,  -42,
  -43,  -44,  -44,  -45,  -46,  -46,  -47,  -48,
  -49,  -49,  -50,  -51,  -51,  -52,  -53,  -54,
  -54,  -55,  -56,  -56,  -57,  -58,  -58,  -59,
  -60,  -61,  -61,  -62,  -63,  -63,  -64,  -65,
  -65,  -66,  -67,  -67,  -68,  -69,  -69,  -70,
  -71,  -71,  -72,  -72,  -73,  -74,  -74,  -75,
  -76,  -76,  -77,  -78,  -78,  -79,  -79,  -80,
  -81,  -81,  -82,  -82,  -83,  -84,  -84,  -85,
  -85,  -86,  -86,  -87,  -88,  -88,  -89,  -89,
  -90,  -90,  -91,  -91,  -92,  -93,  -93,  -94,
  -94,  -95,  -95,  -96,  -96,  -97,  -97,  -98,
  -98,  -99,  -99, -100, -100, -101, -101, -102,
 -102, -102, -103, -103, -104, -104, -105, -105,
 -106, -106, -106, -107, -107, -108, -108, -109,
 -109, -109, -110, -110, -111, -111, -111, -112,
 -112, -112, -113, -113, -113, -114, -114, -114,
 -115, -115, -115, -116, -116, -116, -117, -117,
 -117, -118, -118, -118, -118, -119, -119, -119,
 -120, -120, -120, -120, -121, -121, -121, -121,
 -122, -122, -122, -122, -122, -123, -123, -123,
 -123, -123, -124, -124, -124, -124, -124, -124,
 -125, -125, -125, -125, -125, -125, -125, -126,
 -126, -126, -126, -126, -126, -126, -126, -126,
 -126, -126, -127, -127, -127, -127, -127, -127,
 -127, -127, -127, -127, -127, -127, -127, -127,
 -127, -127, -127, -127, -127, -127, -127, -127,
 -127, -127, -127, -127, -127, -127, -127, -126,
 -126, -126, -126, -126, -126, -126, -126, -126,
 -126, -126, -125, -125, -125, -125, -125, -125,
 -125, -124, -124, -124, -124, -124, -124, -123,
 -123, -123, -123, -123, -122, -122, -122, -122,
 -122, -121, -121, -121, -121, -120, -120, -120,
 -120, -119, -119, -119, -118, -118, -118, -118,
 -117, -117, -117, -116, -116, -116, -115, -115,
 -115, -114, -114, -114, -113, -113, -113, -112,
 -112, -112, -111, -111, -111, -110, -110, -109,
 -109, -109, -108, -108, -107, -107, -106, -106,
 -106, -105, -105, -104, -104, -103, -103, -102,
 -102, -102, -101, -101, -100, -100,  -99,  -99,
  -98,  -98,  -97,  -97,  -96,  -96,  -95,  -95,
  -94,  -94,  -93,  -93,  -92,  -91,  -91,  -90,
  -90,  -89,  -89,  -88,  -88,  -87,  -86,  -86,
  -85,  -85,  -84,  -84,  -83,  -82,  -82,  -81,
  -81,  -80,  -79,  -79,  -78,  -78,  -77,  -76,
  -76,  -75,  -74,  -74,  -73,  -72,  -72,  -71,
  -71,  -70,  -69,  -69,  -68,  -67,  -67,  -66,
  -65,  -65,  -64,  -63,  -63,  -62,  -61,  -61,
  -60,  -59,  -58,  -58,  -57,  -56,  -56,  -55,
  -54,  -54,  -53,  -52,  -51,  -51,  -50,  -49,
  -49,  -48,  -47,  -46,  -46,  -45,  -44,  -44,
  -43,  -42,  -41,  -41,  -40,  -39,  -38,  -38,
  -37,  -36,  -35,  -35,  -34,  -33,  -32,  -32,
  -31,  -30,  -29,  -29,  -28,  -27,  -26,  -26,
  -25,  -24,  -23,  -22,  -22,  -21,  -20,  -19,
  -19,  -18,  -17,  -16,  -16,  -15,  -14,  -13,
  -12,  -12,  -11,  -10,   -9,   -9,   -8,   -7,
   -6,   -5,   -5,   -4,   -3,   -2,   -2,   -1
};
// Convert frequency to in Hz to a phase increment for the sample frequency
unsigned long sineFrequency2phaseinc(float freq, float sampleFreq) {
   return ((SINETABLESIZE*(1L<<PHASEFRACBITS)*freq)/sampleFreq);
}
// Get the sine value for a given phase.
// It will take care of the necessary truncation to fit in the table.
char getSineValue(unsigned long phase) {
   unsigned long index = (phase >> PHASEFRACBITS) & SINETABLEMASK;
   return (pgm_read_byte_near(sinetable+index));
}
//===============================================
// Generate other waveforms
//===============================================
// Use the 256 bit value of the index to calculate the wave
const unsigned int  TABLESIZE = 256;
const unsigned long TABLEMASK = 0x0FF; // only least significant 8 bits used
// Convert frequency to in Hz to a phase increment for the sample frequency
unsigned long frequency2phaseinc(float freq, float sampleFreq) {
   return ((TABLESIZE*(1L<<PHASEFRACBITS)*freq)/sampleFreq);
}
// Get the sine value for a given phase.
// It will take care of the necessary truncation to fit in the table.
word getPhaseValue(unsigned long phase) {
   return ((phase >> PHASEFRACBITS) & TABLEMASK);
}
//===============================================
// Functions for setting wave parameters
//===============================================
// In the functions onum is the number of the oscillator to adjust
void adjustPhaseIncrement(unsigned int onum) {
   unsigned long phaseinc = 0;
   switch (oscarr[onum].type) {
     case SINE:
       phaseinc = sineFrequency2phaseinc(oscarr[onum].frequency, SAMPLEFREQ);
       break;
     case PULSE:
     case TRIANGLE:
     case SAWTOOTH:
       phaseinc = frequency2phaseinc(oscarr[onum].frequency, SAMPLEFREQ);
       break;
     default: // does not occur
       phaseinc = 0;
   }
   oscarr[onum].phase_increment = phaseinc;
//   oscarr[onum].phase = 0; // reset wave
}
void setFrequency(unsigned int onum, float f) {
   oscarr[onum].frequency = f;
   adjustPhaseIncrement(onum);
}
void setAmplitude(unsigned int onum, float a) {
   oscarr[onum].amplitude = (long)(max(0.0,min(1.0,a))*(float)AMPLNORM);
}
void setType(unsigned int onum, Osctype t) {
   oscarr[onum].type = t;
   adjustPhaseIncrement(onum);
}
void setDutyCycle(unsigned int onum, float d) {
}
void setSync(unsigned int onum, boolean b) {
   oscarr[onum].sync = b;
}
void reset() {
   for (int i=0; i<NUMOSC; i++) {
      if (oscarr[i].sync) {
         oscarr[i].phase = 0;
      }
   }
}
void initialize() {
   for (int i=0; i<NUMOSC; i++) {
      oscarr[i].frequency       = 0;
      oscarr[i].phase           = 0;
      oscarr[i].phase_increment = 0;
      oscarr[i].amplitude       = AMPLNORM;
      oscarr[i].type            = SINE;
      oscarr[i].sync            = false;
   }
   DDRD = B11111111;    // PORT D as DAC output    
   // sync pulse input
   pinMode(SYNCPIN, INPUT);
   digitalWrite(SYNCPIN, HIGH); // enable pull-up resistor to prevent false triggering when not connected
   PCintPort::attachInterrupt(SYNCPIN, &reset_wave, FALLING);
   // Set Timer 2 to generate dummy sync pulse on D11
//   pinMode(11, OUTPUT);
//   TCCR2A = _BV(COM2A1) | _BV(WGM20);          // phase-correct pwm, only comparator A
//   TCCR2B = _BV(CS22) | _BV(CS21) | _BV(CS20); // prescaler = 1024
//   OCR2A = 250;
//   OCR2B = 0;
   // output data
   outputvalue     = 0;
   errortimeout    = 0; // no error
   // initialize the digital error LED pin as an output.
   pinMode(ERRORLED, OUTPUT);     
   digitalWrite(ERRORLED, LOW); // turn the LED off by making the voltage LOW
}
//===============================================
// The wave generating interrupt routine
//===============================================
void generateWave() {
    // First output value to minimize jitter due to calculation time
    DACOUT = (byte)(outputvalue + (int)127); // shift wave to make 127 the mean value
    long oval = 0;
    long tmpval;
    unsigned long phase;
    // Check if waves have to be restarted
    if (restart_wave) {
       reset();
       restart_wave = false;
    }
    // Calculate new output value
    for (int i=0; i<NUMOSC; i++) {
       oscarr[i].phase += oscarr[i].phase_increment; // phase of next sample
       switch (oscarr[i].type) {
       case SINE:
          tmpval = getSineValue(oscarr[i].phase);
          break;
       case PULSE:
          phase = getPhaseValue(oscarr[i].phase);
          if (phase <= 127) {
             tmpval = 127;
          }
          else {
             tmpval = -127;
          }
          break;
       case TRIANGLE:
          // TODO: make this also one step per phase (512 size table)
          phase = getPhaseValue(oscarr[i].phase);
          if (phase <= 127) {
             tmpval = phase;
          }
          else {
             tmpval = 255 - phase;
          }
          tmpval = (tmpval << 1) - 127; // center around 0
          break;
       case SAWTOOTH:
          phase = getPhaseValue(oscarr[i].phase);
          tmpval = phase - 127; // center around 0
          break;
       default: // does not occur
          ; // do nothing
       }
       oval = oval + (tmpval * oscarr[i].amplitude);
    }
    oval = oval / AMPLNORM;
    if (abs(oval) > AMPLITUDEMAX) {
       // peaking wave
       errortimeout = SAMPLEFREQ; // show peak on LED for 1 second
       digitalWrite(ERRORLED, HIGH);
    }
    else if (errortimeout > 0) {
       errortimeout--;
       if (errortimeout == 0) {
          digitalWrite(ERRORLED, LOW);
       }
    }
    outputvalue = oval;
}
//===============================================
// Interrupt routine that resets oscillators
//===============================================
//volatile int ai=0;
void reset_wave() {
  restart_wave = true;
  TCNT1 = 0;                	
}
//===============================================
// Setup and loop functions
//===============================================
void setup() {
  
   pinMode(19, OUTPUT);
   // N.B Serial communication uses D0 and D1, so it can't be used!
   noInterrupts(); // no interrupts while setting up
   // The set samplefrequency can differ from the desired value!
   Timer1.initialize((unsigned long)(1000000L / SAMPLEFREQ)); // parameter in microseconds
   Timer1.attachInterrupt(generateWave);
   initialize();
   setType(0, TRIANGLE);
   setFrequency(0, 61.0);
   setAmplitude(0, 0.6);
   setSync(0, false);
   setType(1, TRIANGLE);
   setFrequency(1, 60.0);
   setAmplitude(1, 0.2);
   setSync(1, false);
   analogReference(INTERNAL);
//   setType(2, SAWTOOTH);
//   setFrequency(2, 0.5);
//   setAmplitude(2, 0.4);
//   setSync(2, false);
   interrupts(); // enable interrupts for wave generation
}

void loop() {
   // nothing to do, all is done in interrupt routine
   int fi = analogRead(FANPIN);
  
   setFrequency(1, 180.0+min(fi/20.0,10.0));
   setAmplitude(0, min(fi/120.0,0.65));
   setFrequency(0, 61.0+min(fi/61.0,2.0));
   setAmplitude(1, min(fi/220.0,0.35));
   
   if(fi > 60){
     digitalWrite(19, HIGH);
   }else if(fi < 1){   
     digitalWrite(19, LOW);
   }
}

